#!/usr/bin/python3

# communityusage.py v1.0
# Authors: Thomas Krenc, Robert Beverly and Georgios Smaragdakis
# E-Mails: tkrenc@nps.edu, rbeverly@nps.edu, g.smaragdakis@tudelft.nl

# This script classifies the community usage behavior of ASes.
# To comment the code, we provide references to the corresponding sections in the paper.
# For more information, see https://www.cmand.org/communityusage

import sys
import time

# labels for community usage behavior
t = {
  'none': 0,
  'tagger': 1,
  'silent': 2,
  'forward': 4,
  'cleaner': 8,
  'undecided1': 16,
  'undecided2': 32
}

# stores counters per AS for as classification
counters = {}

# hold the entire input data in buffer for multiple iterations
buffer = {}
buffer_keys = []
buffer_values = []

# threshold required for classification
thrs_global = 0.99

thrs_tagger = thrs_global
thrs_silent = thrs_global
thrs_forward = thrs_global
thrs_cleaner = thrs_global

# minimum count required for classification
mincnt_global = 1

mincnt_tagger = mincnt_global
mincnt_silent = mincnt_global
mincnt_forward = mincnt_global
mincnt_cleaner = mincnt_global

# maximum path length / iterations
max_pathlength = 10

# start algorithm and print classification results
def main():

  readInput()

  # iterate over input data (see Section 5.6 "Algorithm")
  for i in range(max_pathlength):
    inferTagging(i)
    inferForwarding(i)

  # print classification results
  for k in counters:
    print(str(k) + " " + getASClass(k))

# read from stdin and store in buffer
def readInput():

  lc=0
  sys.stderr.write("[" + time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()) + "] readInput(): begin ")
  sys.stderr.flush()
  for line in sys.stdin:
    lc=lc+1
    if lc % 10000000 == 0:
      sys.stderr.write(".")
      sys.stderr.flush()
    _line = line.rstrip().split("|")
    # if number of occurrences is not supplied, default to 1
    if len(_line) == 2: _line.append('1')
    _asp, _cos, _count = _line
    asp = _asp.split(' ') if _asp != '' else []
    # initialize counters for each AS
    for a in asp:
      if a not in counters:
        counters[a] = { t['tagger']: 0, t['silent']: 0, t['forward']: 0, t['cleaner']: 0 }
    key = _asp + '|' + _cos
    # fill buffer
    if key not in buffer:
      buffer[key] = 0
    buffer[key] += int(_count)
  lc=0
  sys.stderr.write("\n[" + time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()) + "] readInput(): sorting ")
  sys.stderr.flush()
  # sort keys for repeatability (see Section 5.2.1 "Race conditions")
  for key in sorted(buffer.keys()):
    lc=lc+1
    if lc % 10000000 == 0:
      sys.stderr.write(".")
      sys.stderr.flush()
    buffer_keys.append(key)
  sys.stderr.write("\n[" + time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()) + "] readInput(): end\n")

  print("SUMMARY", end='')
  print(" total path/comm pairs:" + str(sum(buffer.values())), end='')
  print(" unique path/comm pairs:" + str(len(buffer)), end='')
  print(" num ASes:" + str(len(counters)))

  print("SUMMARY", end='')
  print(" tagger: " + str(thrs_tagger), end='')
  print(" silent: " + str(thrs_silent), end='')
  print(" forward: " + str(thrs_forward), end='')
  print(" cleaner: " + str(thrs_cleaner))

def inferTagging(x):

  lc=0

  for key in buffer_keys: # OUTER LOOP

    lc=lc+1
    if lc % 10000000 == 0:
      sys.stderr.write("[" + time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()) + "] inferTagging(), path index: " + str(x) + ", read line: " + str(lc) + "\n")

    _cnt_key = buffer[key]

    # key
    _key = key.split('|')
    # as path / sequence
    asp = _key[0].split(' ') if _key[0] != "" else []
    # community set
    com = _key[1].split(' ') if _key[1] != "" else []

    # skip if path is not long enough
    if len(asp) < (1+x):
      continue


#<-- CONDITION 1: all upstream ASes must be ingore (see Section 5.2 "Conditions")
    _f_continue_outer_loop = False
    if x > 0:

      # all asp[i] with i < x must be of type forward
      for i in range(0, x):

        if not ( isForward(asp[i]) ):
          _f_continue_outer_loop = True
          break

    if _f_continue_outer_loop:
      continue
# CONDITION 1 -->

# in first iteration (x=0) an _emtpy comm.string_ as indication for a silent peer AS (see Section 5.3. "Counting community usage")

    if ( asp[x] in com ):
      counters[asp[x]][t['tagger']] += _cnt_key
    else:
      counters[asp[x]][t['silent']] += _cnt_key



def inferForwarding(x):

  lc=0

  for key in buffer_keys: # OUTER LOOP

    lc=lc+1
    if lc % 10000000 == 0:
      sys.stderr.write("[" + time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()) + "] inferForwarding(), path index: " + str(x) + ", read line: " + str(lc) + "\n")

    _cnt_key = buffer[key]

    # key
    _key = key.split('|')
    # as path / sequence
    asp = _key[0].split(' ') if _key[0] != "" else []
    # community set
    com = _key[1].split(' ') if _key[1] != "" else []

    # skip if path is not long enough
    if len(asp) < (2+x):
      continue

#<-- CONDITION 1: all upstream ASes must be ingore (see Section 5.2 "Conditions")
    _f_continue_outer_loop = False
    if x > 0:

      # all asp[i] with i < x must be of type forward
      for i in range(0, x):

        # if a single asp[i] is not forward, abort
        if not isForward(asp[i]):
          _f_continue_outer_loop = True
          break

    if _f_continue_outer_loop:
      continue
# CONDITION 1 -->

#<-- CONDITION 2: there must exist a visible downstream tagger (see Section 5.2 "Conditions")
    _visible_tagger = []
    if x+1 < len(asp):

      for i in range(x+1, len(asp)):

        if isTagger(asp[i]):
          _visible_tagger.append(asp[i])

        if not isForward(asp[i]):
          break

    if len(_visible_tagger) == 0:
      continue
# CONDITION 2 -->

    _comrest = [c for c in com if c in _visible_tagger]

    # if communities from visible downstream taggers exist, asp[x] is a forward AS (see Section 5.3. "Counting community usage")
    if len(_comrest) > 0:
      counters[asp[x]][t['forward']] += _cnt_key
    else:
      counters[asp[x]][t['cleaner']] += _cnt_key


### functions to determine the tagging and forwarding behavior based on counters (see Section 5.3. "Counting community usage")

def isTagger(asn):

  _t, _s = counters[asn][t['tagger']], counters[asn][t['silent']]

  return False if _t + _s == 0 else (True if _t / (_t + _s) >= thrs_tagger else False)

def isSilent(asn):

  _t, _s = counters[asn][t['tagger']], counters[asn][t['silent']]

  return False if _t + _s == 0 else (True if _s / (_t + _s) >= thrs_silent else False)

def isForward(asn):

  _f, _c = counters[asn][t['forward']], counters[asn][t['cleaner']]

  return False if _f + _c == 0 else (True if _f / (_f + _c) >= thrs_forward else False)

def isCleaner(asn):

  _f, _c = counters[asn][t['forward']], counters[asn][t['cleaner']]

  return False if _f + _c == 0 else (True if _c / (_f + _c) >= thrs_cleaner else False)

# create label to indicate the inferred tagging and forwarding behavior (see Section 5.5 "Classification")
def getASClass(asn):

  _t, _s, _f, _c = counters[asn][t['tagger']], counters[asn][t['silent']], counters[asn][t['forward']], counters[asn][t['cleaner']]

  _label = ''

  if _t + _s > 0:
    if isTagger(asn):
      _label += 't'
    elif isSilent(asn):
      _label += 's'
    else:
      _label += 'u'
  else:
    _label += 'n'

  if _f + _c > 0:
    if isForward(asn):
      _label += 'f'
    elif isCleaner(asn):
      _label += 'c'
    else:
      _label += 'u'
  else:
    _label += 'n'

  return _label


main()
